@@ -84,7 +84,7 @@ class Ec2:
 
         response = self.__ec2_best_instance.get_best_instance_types(get_best_instance_types_input)
 
-        return response[0]
+        return response
 
     def is_instance_storage_supported_for_instance_type(self, instance_type):
         return self.__ec2_best_instance.is_instance_storage_supported_for_instance_type(instance_type)
