@@ -11,6 +11,7 @@ from .dynamo_db import DynamoDb
 from .run_status import RunStatus
 from .fsx import Fsx
 from .ec2 import Ec2
+from .exception.InstanceDeployingTimeoutException import InstanceDeployingTimeoutException
 
 
 class ContainerLauncher:
@@ -94,9 +95,15 @@ class ContainerLauncher:
         iam_ec2_policy_documents = None
         iam_ec2_managed_policy_arns = []
         max_spot_interruption_frequency = 10
+        max_spot_instance_type_candidates = 3
+        spot_allocation_strategy = 'capacity-optimized-prioritized'
         if container_input.get('spot') is not None:
             if container_input['spot'].get('max_interruption_frequency') is not None:
                 max_spot_interruption_frequency = container_input['spot']['max_interruption_frequency']
+            if container_input['spot'].get('max_instance_type_candidates') is not None:
+                max_spot_instance_type_candidates = container_input['spot']['max_instance_type_candidates']
+            if container_input['spot'].get('allocation_strategy') is not None:
+                spot_allocation_strategy = container_input['spot']['allocation_strategy']
         if container_input.get('iam') is not None:
             if container_input['iam'].get('ecs_task') is not None:
                 if container_input['iam']['ecs_task'].get('policy_documents'):
@@ -131,6 +138,7 @@ class ContainerLauncher:
 
             if container_input.get('instance_type'):
                 instance_type = container_input['instance_type']
+                ec2_instances = [{'instance_type': instance_type}]
             else:
                 ec2 = Ec2({
                     'region': self.__region
@@ -147,7 +155,8 @@ class ContainerLauncher:
                     if len(ssd_instance_stores) > 0:
                         get_optimal_instance_type_input['is_instance_storage_supported'] = True
 
-                ec2_instance = ec2.get_optimal_instance_type(get_optimal_instance_type_input)
+                ec2_instances = ec2.get_optimal_instance_type(get_optimal_instance_type_input)
+                ec2_instance = ec2_instances[0]
                 self.logger.debug(f'The optimal instance type is {str(ec2_instance)}')
                 instance_type = ec2_instance['instance_type']
 
@@ -181,17 +190,27 @@ class ContainerLauncher:
                 'subnet_ids': subnet_ids,
                 'azs': self.__get_azs_by_subnet_ids(subnet_ids),
                 'account_id': account_id,
-                'entry_id': entry_id
+                'entry_id': entry_id,
+                'ec2_instances': ec2_instances,
+                'usage_class': usage_class,
+                'max_spot_instance_type_candidates': max_spot_instance_type_candidates,
+                'spot_allocation_strategy': spot_allocation_strategy
             })
             if container_input.get('options') and container_input['options'].get('scaling_enabled') and \
                     container_input['options']['scaling_enabled'] is True:
-                self.__add_scale_in_policy({})
+                self.__add_scale_in_policy(
+                    container_input['options']['scaling_options'] if container_input['options'].get(
+                        'scaling_options') is not None else {})
             self.__create_capacity_provider()
             self.__create_ecs_cluster()
             self.__wait_for_instance_deploying(is_spot)
             self.__wait_for_instance_registration_in_ecs()
 
             memory = self.__get_available_memory()
+            instance_type = self.__get_asg_instance_type()
+
+            self.logger.info(
+                f'CPU: {cpu}, memory: {memory}, Instance type: {instance_type}, Usage class: {usage_class}, Docker: {docker_image}')
 
             ecs_task_definition_data = self.__get_ecs_task_definition_data(
                 container_input=container_input,
@@ -201,9 +220,6 @@ class ContainerLauncher:
                 commands=commands
             )
 
-            self.logger.info(
-                f'CPU: {cpu}, Memory: {memory}, Instance type: {instance_type}, Usage class: {usage_class}, Docker: {docker_image}')
-
             self.__create_ecs_task_definition(ecs_task_definition_data, instance_type)
 
             self.__dynamodb.update_state(entry_id, self.__resources)
@@ -227,7 +243,21 @@ class ContainerLauncher:
                 'entry_id': entry_id,
                 'logs': logs
             }
-
+        except InstanceDeployingTimeoutException as e:
+            code = 'InstanceDeployingTimeoutException'
+            self.logger.error(self.__resources)
+            self.logger.error(e)
+            self.__dynamodb.update_status(entry_id, RunStatus.FAILED.name)
+            self.__dynamodb.add_reason(entry_id, code)
+            self.logger.info('Destroying resources...')
+            self.__destroy()
+            return {
+                'entry_id': entry_id,
+                'error': {
+                    'code': code,
+                    'message': str(e)
+                }
+            }
         except Exception as e:
             self.logger.error(self.__resources)
             self.logger.error(e)
@@ -247,10 +277,15 @@ class ContainerLauncher:
         run_status = entry['status']
 
         if run_status != RunStatus.IN_PROGRESS.name:
-            return {
+            response = {
                 'status': run_status
             }
 
+            if entry.get('reason') is not None:
+                response['response'] = entry['reason']
+
+            return response
+
         response = self.clients['ecs'].describe_tasks(
             cluster=entry['ecs_cluster_name'],
             tasks=[entry['task_arn']]
@@ -607,6 +642,41 @@ class ContainerLauncher:
         availability_zones = asg_opts['azs']
         account_id = asg_opts['account_id']
         entry_id = asg_opts['entry_id']
+        ec2_instances = asg_opts['ec2_instances']
+        usage_class = asg_opts['usage_class']
+        max_spot_instance_type_candidates = asg_opts['max_spot_instance_type_candidates']
+        spot_allocation_strategy = asg_opts['spot_allocation_strategy']
+
+        launch_template = {}
+        mixed_instances_policy = {}
+
+        if usage_class == 'spot':
+            overrides = []
+            for i, ec2_instance in enumerate(ec2_instances):
+                overrides.append({
+                    'InstanceType': ec2_instance['instance_type']
+                })
+                if i == max_spot_instance_type_candidates - 1:
+                    break
+
+            mixed_instances_policy = {
+                'LaunchTemplate': {
+                    'LaunchTemplateSpecification': {
+                        'LaunchTemplateName': self.__resources['launch_template_name'],
+                        'Version': '$Latest'
+                    },
+                    'Overrides': overrides
+                },
+                'InstancesDistribution': {
+                    'SpotAllocationStrategy': spot_allocation_strategy
+                }
+            }
+            self.logger.info(
+                f'Instance type candidates: {str(overrides)}, allocation strategy: {spot_allocation_strategy}')
+        else:
+            launch_template = {
+                'LaunchTemplateName': self.__resources['launch_template_name'],
+            }
 
         tags = list(map(lambda tag: {
             'Key': tag['Key'],
@@ -622,9 +692,7 @@ class ContainerLauncher:
 
         self.clients['autoscaling'].create_auto_scaling_group(
             AutoScalingGroupName=asg_name,
-            LaunchTemplate={
-                'LaunchTemplateName': self.__resources['launch_template_name'],
-            },
+            LaunchTemplate=launch_template,
             MinSize=0,
             MaxSize=1,
             DesiredCapacity=1,
@@ -632,7 +700,7 @@ class ContainerLauncher:
             AvailabilityZones=availability_zones,
             HealthCheckType='EC2',
             HealthCheckGracePeriod=0,
-
+            MixedInstancesPolicy=mixed_instances_policy,
             VPCZoneIdentifier=vpc_zone_identifier,
             TerminationPolicies=[
                 'Default',
@@ -810,6 +878,7 @@ class ContainerLauncher:
 
     def __add_scale_in_policy(self, scaling_opts):
         asg_name = self.__resources['asg_name']
+        period = scaling_opts.get('period', 300)
         evaluation_periods = scaling_opts.get('evaluation_periods', 36)
         threshold = scaling_opts.get('threshold', 1.0)
         name_suffix = self.__options['name_suffix']
@@ -837,7 +906,7 @@ class ContainerLauncher:
                     'Value': asg_name
                 },
             ],
-            Period=300,
+            Period=period,
             Unit='Percent',
             EvaluationPeriods=evaluation_periods,
             Threshold=threshold,
@@ -958,16 +1027,17 @@ class ContainerLauncher:
 
         start_time = datetime.datetime.now()
 
-        self.logger.info('Waiting for an EC2 instance deploying')
-
         timeout_in_minutes = self.__SPOT_INSTANCE_DEPLOYING_TIMEOUT_IN_MINUTES if is_spot else self.__EC2_INSTANCE_DEPLOYING_TIMEOUT_IN_MINUTES
 
+        self.logger.info(f'Waiting for an EC2 instance deploying ({timeout_in_minutes} minutes)')
+
         while True:
             now_time = datetime.datetime.now()
             diff = now_time - start_time
             diff_minutes = diff.total_seconds() / 60
             if diff_minutes > timeout_in_minutes:
-                raise Exception(f'EC2 instance deploying timeout ({timeout_in_minutes} minutes)')
+                raise InstanceDeployingTimeoutException(
+                    f'EC2 instance deploying timeout ({timeout_in_minutes} minutes)')
             if len(instances) > 0:
                 self.logger.debug('The EC2 instance ID: ' + instances[0]['InstanceId'])
                 response = self.clients['ec2'].describe_instance_status(InstanceIds=[instances[0]['InstanceId']])
@@ -979,7 +1049,7 @@ class ContainerLauncher:
                         break
             time.sleep(10)
             instances = self.__get_asg_instances()
-            self.logger.debug('Waiting for an EC2 instance deploying')
+            self.logger.debug(f'Waiting for an EC2 instance deploying')
 
         if len(instances) > 0:
             instance_id = instances[0]['InstanceId']
@@ -1079,21 +1149,21 @@ mount /dev/md0 %s
                 mount_name = file_system['LustreConfiguration']['MountName']
                 install_fsx_agent_cmd = 'amazon-linux-extras install -y lustre2.10'
                 user_data = f'{user_data}\n{install_fsx_agent_cmd}\nmkdir {fsx_dir}\nmount -t lustre -o noatime,flock {dns_name}@tcp:/{mount_name} {fsx_dir}'
-        block_devices = None
-        if storages is not None:
-            ebs_storages = [storage for storage in storages if storage['type'] == 'ebs']
-            block_devices = []
-            for ebs_storage in ebs_storages:
-                block_device = {
-                    'DeviceName': ebs_storage['device_name'],
-                    'Ebs': {
-                        'Encrypted': False,
-                        'DeleteOnTermination': True,
-                        'VolumeSize': ebs_storage['volume_size'],
-                        'VolumeType': ebs_storage['volume_type']  # 'standard'|'io1'|'io2'|'gp2'|'sc1'|'st1'|'gp3',
-                    }
-                }
-                block_devices.append(block_device)
+        # block_devices = None
+        # if storages is not None:
+        #    ebs_storages = [storage for storage in storages if storage['type'] == 'ebs']
+        #    block_devices = []
+        #    for ebs_storage in ebs_storages:
+        #        block_device = {
+        #            'DeviceName': ebs_storage['device_name'],
+        #            'Ebs': {
+        #                'Encrypted': False,
+        #                'DeleteOnTermination': True,
+        #                'VolumeSize': ebs_storage['volume_size'],
+        #                'VolumeType': ebs_storage['volume_type']  # 'standard'|'io1'|'io2'|'gp2'|'sc1'|'st1'|'gp3',
+        #            }
+        #        }
+        #        block_devices.append(block_device)
 
         output_logs_command = 'exec > >(tee /var/log/user-data.log|logger -t user-data -s 2>/dev/console) 2>&1'
         user_data = f'#!/bin/bash\n{install_ecs_agent_cmd}\n{output_logs_command}\n{user_data}'
@@ -1118,16 +1188,16 @@ mount /dev/md0 %s
             ]
         }
 
-        if block_devices is not None and len(block_devices) > 0:
-            launch_template_data['BlockDeviceMappings'] = block_devices
+        # if block_devices is not None and len(block_devices) > 0:
+        #    launch_template_data['BlockDeviceMappings'] = block_devices
 
         if ec2_sg_ids is not None and len(ec2_sg_ids) > 0:
             launch_template_data['SecurityGroupIds'] = ec2_sg_ids
 
-        if is_spot:
-            launch_template_data['InstanceMarketOptions'] = {
-                'MarketType': 'spot'
-            }
+        # if is_spot:
+        #    launch_template_data['InstanceMarketOptions'] = {
+        #        'MarketType': 'spot'
+        #    }
 
         self.clients['ec2'].create_launch_template(
             LaunchTemplateName=launch_template_name,
@@ -1211,6 +1281,18 @@ mount /dev/md0 %s
 
         return ecs_task_definition
 
+    def __get_asg_instance_type(self):
+        response = self.clients['autoscaling'].describe_auto_scaling_groups(
+            AutoScalingGroupNames=[self.__resources['asg_name']]
+        )
+
+        instances = response['AutoScalingGroups'][0]['Instances']
+
+        if len(instances) > 0:
+            return instances[0]['InstanceType']
+
+        return ''
+
     @staticmethod
     def __filter_not_terminated_instances(ec2_instance):
         if ec2_instance.get('State') is not None and ec2_instance['State']['Name'] == 'terminated':
